/*
   Lagging Fibonacci Pseudo Random Generator for integer types.

   The design of this template class is partially based on
   boost/random/lagged_fibonacci.hpp of the Boost library, see
   http://www.boost.org which is copyrighted by Jens Maurer 2000-2001
   and freely distributable under the Boost Software License:
   http://www.boost.org/LICENSE_1_0.txt.

   An instance of LFGInt generates unsigned integers of the type
   UnsignedInteger out of the range [0..2^wordWidth-1] of the
   lagged Fibonacci sequence

      x_n = (x_{n-p} + x_{n-q}) mod 2^wordWidth

   where p > q. The p seed values are extracted from the generator
   that is passed to the constructor using the () operator.

   Andreas F. Borchert
*/

#ifndef LFG_INT_H
#define LFG_INT_H

template<
   typename UnsignedInteger,
   unsigned int wordWidth,
   unsigned int p, // p > q
   unsigned int q
> class LFGInt {
   public:
      typedef UnsignedInteger result_type;
      enum { long_lag = p };
      enum { short_lag = q };
      enum { lag_diff = p-q };

      template<typename Generator>LFGInt(Generator& gen) :
	    index(p), mask(get_mask()) {
	 for (int i = 0; i < long_lag; ++i) {
	    values[i] = gen() & mask;
	 }
      }

      UnsignedInteger operator()() {
	 if (index >= long_lag) {
	    fill();
	 }
	 return values[index++];
      }

   private:
      UnsignedInteger values[p];
      const UnsignedInteger mask;
      unsigned int index;

      void fill() {
	 for (int i = 0; i < short_lag; ++i) {
	    values[i] = (values[i] + values[i + lag_diff]) & mask;
	 }
	 for (int i = short_lag; i < long_lag; ++i) {
	    values[i] = (values[i] + values[i - short_lag]) & mask;
	 }
	 index = 0;
      }

      static UnsignedInteger get_mask() {
	 UnsignedInteger m = 0;
	 for (int i = 0; i < wordWidth; ++i) {
	    m |= 1 << i;
	 }
	 return m;
      }
};

#endif
